@@ -3,7 +3,7 @@
  *
  * Tcl command execution trace command.
  *-----------------------------------------------------------------------------
- * Copyright 1991-1996 Karl Lehenbauer and Mark Diekhans.
+ * Copyright 1991-1997 Karl Lehenbauer and Mark Diekhans.
  *
  * Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -12,7 +12,7 @@
  * software for any purpose.  It is provided "as is" without express or
  * implied warranty.
  *-----------------------------------------------------------------------------
- * $Id: tclXdebug.c,v 5.3 1996/02/12 18:15:36 markd Exp $
+ * $Id: tclXdebug.c,v 8.9 1997/08/17 05:31:08 markd Exp $
  *-----------------------------------------------------------------------------
  */
 
@@ -25,15 +25,17 @@
 #define CMD_TRUNCATE_SIZE 60
 
 typedef struct traceInfo_t {
-    Tcl_Interp *interp;
-    Tcl_Trace   traceHolder;
-    int         inTrace;
-    int         noEval;
-    int         noTruncate;
-    int         procCalls;
-    int         depth;
-    char       *callback;
-    Tcl_Channel channel;
+    Tcl_Interp       *interp;
+    Tcl_Trace         traceId;
+    int               inTrace;
+    int               noEval;
+    int               noTruncate;
+    int               procCalls;
+    int               depth;
+    char             *callback;
+    Tcl_Obj          *errorStatePtr;
+    Tcl_AsyncHandler  errorAsyncHandler;
+    Tcl_Channel       channel;
     } traceInfo_t, *traceInfo_pt;
 
 /*
@@ -61,6 +63,11 @@ TraceCode  _ANSI_ARGS_((traceInfo_pt infoPtr,
                         int          argc,
                         char       **argv));
 
+static int
+TraceCallbackErrorHandler _ANSI_ARGS_((ClientData  clientData,
+                                       Tcl_Interp *interp,
+                                       int         code));
+
 static void
 TraceCallBack _ANSI_ARGS_((Tcl_Interp   *interp,
                            traceInfo_pt  infoPtr,
@@ -79,13 +86,18 @@ CmdTraceRoutine _ANSI_ARGS_((ClientData    clientData,
                              int           argc,
                              char        **argv));
 
+static int
+TclX_CmdtraceObjCmd _ANSI_ARGS_((ClientData clientData, 
+                                 Tcl_Interp *interp,
+                                 int objc,
+                                 Tcl_Obj *CONST objv[]));
+
 static void
 DebugCleanUp _ANSI_ARGS_((ClientData  clientData,
                           Tcl_Interp *interp));
 
 
-/*
- *-----------------------------------------------------------------------------
+/*-----------------------------------------------------------------------------
  * TraceDelete --
  *
  *   Delete the trace if active, reseting the structure.
@@ -96,19 +108,22 @@ TraceDelete (interp, infoPtr)
     Tcl_Interp   *interp;
     traceInfo_pt  infoPtr;
 {
-    if (infoPtr->traceHolder != NULL) {
-        Tcl_DeleteTrace (interp, infoPtr->traceHolder);
+    if (infoPtr->traceId != NULL) {
+        Tcl_DeleteTrace (interp, infoPtr->traceId);
         infoPtr->depth = 0;
-        infoPtr->traceHolder = NULL;
+        infoPtr->traceId = NULL;
         if (infoPtr->callback != NULL) {
             ckfree (infoPtr->callback);
             infoPtr->callback = NULL;
         }
     }
+    if (infoPtr->errorAsyncHandler != NULL) {
+        Tcl_AsyncDelete (infoPtr->errorAsyncHandler);
+        infoPtr->errorAsyncHandler = NULL;
+    }
 }
 
-/*
- *-----------------------------------------------------------------------------
+/*-----------------------------------------------------------------------------
  * PrintStr --
  *
  *     Print an string, truncating it to the specified number of characters.
@@ -133,19 +148,18 @@ PrintStr (channel, string, numChars, quoted)
             Tcl_Write (channel, &(string [idx]), 1);
         }
     }
-    if (numChars < strlen (string))
+    if (numChars < (int) strlen (string))
         Tcl_Write (channel, "...", 3);
     if (quoted) 
         Tcl_Write (channel, "}", 1);
 }
 
-/*
- *-----------------------------------------------------------------------------
+/*-----------------------------------------------------------------------------
  * PrintArg --
  *
  *   Print an argument string, truncating and adding "..." if its longer
  * then ARG_TRUNCATE_SIZE.  If the string contains white spaces, quote
- * it with angle brackets.
+ * it with braces.
  *-----------------------------------------------------------------------------
  */
 static void
@@ -173,8 +187,7 @@ PrintArg (channel, argStr, noTruncate)
     PrintStr (channel, argStr, printLen, quoted);
 }
 
-/*
- *-----------------------------------------------------------------------------
+/*-----------------------------------------------------------------------------
  * TraceCode --
  *
  *   Print out a trace of a code line.  Level is used for indenting
@@ -219,9 +232,39 @@ TraceCode (infoPtr, level, command, argc, argv)
     TclX_WriteNL (infoPtr->channel);
     Tcl_Flush (infoPtr->channel);
 }
+
 
-/*
+/*-----------------------------------------------------------------------------
+ * TraceCallbackErrorHandler --
+ *
+ *   Async handler that processes an callback error.  Generates either an
+ * immediate or background error.
  *-----------------------------------------------------------------------------
+ */
+static int
+TraceCallbackErrorHandler (clientData, interp, code)
+    ClientData  clientData;
+    Tcl_Interp *interp;
+    int         code;
+{
+    traceInfo_pt infoPtr = (traceInfo_pt) clientData;
+
+    /*
+     * Put back error message and state.  If not interp passed in, the error
+     * is handled in the background.
+     */
+    TclX_RestoreResultErrorInfo (infoPtr->interp, infoPtr->errorStatePtr);
+    infoPtr->errorStatePtr = NULL;
+    if (interp == NULL) {
+        Tcl_BackgroundError (infoPtr->interp);
+    }    
+    
+    TraceDelete (interp, infoPtr);
+
+    return TCL_ERROR;
+}
+
+/*-----------------------------------------------------------------------------
  * TraceCallBack --
  *
  *   Build and call a callback for the command that was just executed. The
@@ -235,8 +278,9 @@ TraceCode (infoPtr, level, command, argc, argv)
  * The code should allow for additional substitution of arguments in future
  * versions (such as a procedure with args as the last argument).  The value
  * of result, errorInfo and errorCode are preserved.  All other state must be
- * preserved by the procedure.  A error will result in information being dumped
- * to stderr and the trace deleted.
+ * preserved by the procedure.  An error will result in an error being flagged
+ * in the control block and async mark being called to handle the error
+ * once the command has completed.
  *-----------------------------------------------------------------------------
  */
 static void
@@ -249,14 +293,12 @@ TraceCallBack (interp, infoPtr, level, command, argc, argv)
     char        **argv;
 {
     Interp       *iPtr = (Interp *) interp;
-    Tcl_DString   callback, resultBuf, errorInfoBuf, errorCodeBuf;
-    char         *cmdList, *errorInfo, *errorCode;
+    Tcl_DString   callback;
+    Tcl_Obj      *saveObjPtr;
+    char         *cmdList;
     char          numBuf [32];
 
     Tcl_DStringInit (&callback);
-    Tcl_DStringInit (&resultBuf);
-    Tcl_DStringInit (&errorInfoBuf);
-    Tcl_DStringInit (&errorCodeBuf);
 
     /*
      * Build the command to evaluate.
@@ -280,61 +322,25 @@ TraceCallBack (interp, infoPtr, level, command, argc, argv)
              iPtr->varFramePtr->level));
     Tcl_DStringAppendElement (&callback, numBuf);
 
-    /*
-     * Save errorInfo and errorCode.
-     */
-    Tcl_DStringGetResult (interp, &resultBuf);
-    errorInfo = Tcl_GetVar (interp, "errorInfo", TCL_GLOBAL_ONLY);
-    if (errorInfo != NULL)
-        Tcl_DStringAppend (&errorInfoBuf, errorInfo, -1);
-    errorCode = Tcl_GetVar (interp, "errorCode", TCL_GLOBAL_ONLY);
-    if (errorCode != NULL)
-        Tcl_DStringAppend (&errorCodeBuf, errorCode, -1);
+    saveObjPtr = TclX_SaveResultErrorInfo (interp);
 
     /*
-     * Evaluate the command.  If an error occurs, dump something to stderr
-     * and delete the trace.  There is no way to return an error at this
-     * point.
+     * Evaluate the command.  If an error occurs, set up the handler to be
+     * called when its possible.
      */
     if (Tcl_Eval (interp, Tcl_DStringValue (&callback)) == TCL_ERROR) {
-        Tcl_Channel stderrChan;
-
-        Tcl_AddErrorInfo (interp, "\n    (\"cmdtrace\" callback command)");
-        
-        stderrChan = TclX_Stderr (interp);
-        if (stderrChan != NULL) {
-            TclX_WriteStr (stderrChan, 
-                           "cmdtrace callback command error: errorCode = ");
-            TclX_WriteStr (stderrChan,
-                           Tcl_GetVar (interp, "errorCode", TCL_GLOBAL_ONLY));
-            TclX_WriteNL (stderrChan);
-            TclX_WriteStr (stderrChan,
-                           Tcl_GetVar (interp, "errorInfo", TCL_GLOBAL_ONLY));
-            TclX_WriteNL (stderrChan);
-            Tcl_Flush (stderrChan);
-        }
-        TraceDelete (interp, infoPtr);
+        Tcl_AddObjErrorInfo (interp, "\n    (\"cmdtrace\" callback command)",
+                             -1);
+        infoPtr->errorStatePtr = TclX_SaveResultErrorInfo (interp);
+        Tcl_AsyncMark (infoPtr->errorAsyncHandler);
     }
 
-    /*
-     * Restore errorInfo and errorCode.
-     */
-    if (errorInfo != NULL)
-        Tcl_SetVar (interp, "errorInfo", Tcl_DStringValue (&errorInfoBuf),
-                    TCL_GLOBAL_ONLY);
-    if (errorCode != NULL)
-        Tcl_SetVar (interp, "errorCode", Tcl_DStringValue (&errorCodeBuf),
-                    TCL_GLOBAL_ONLY);
-    Tcl_DStringResult (interp, &resultBuf);
+    TclX_RestoreResultErrorInfo (interp, saveObjPtr);
 
     Tcl_DStringFree (&callback);
-    Tcl_DStringFree (&resultBuf);
-    Tcl_DStringFree (&errorInfoBuf);
-    Tcl_DStringFree (&errorCodeBuf);
 }
 
-/*
- *-----------------------------------------------------------------------------
+/*-----------------------------------------------------------------------------
  * CmdTraceRoutine --
  *
  *  Routine called by Tcl_Eval to trace a command.
@@ -356,8 +362,12 @@ CmdTraceRoutine (clientData, interp, level, command, cmdProc, cmdClientData,
     traceInfo_pt  infoPtr = (traceInfo_pt) clientData;
     int           procLevel;
 
-    if (infoPtr->inTrace)
+    /*
+     * If we are in an error.  
+     */
+    if (infoPtr->inTrace || (infoPtr->errorStatePtr != NULL)) {
         return;
+    }
     infoPtr->inTrace = TRUE;
 
     if (infoPtr->procCalls) {
@@ -380,9 +390,8 @@ CmdTraceRoutine (clientData, interp, level, command, cmdProc, cmdClientData,
     infoPtr->inTrace = FALSE;
 }
 
-/*
- *-----------------------------------------------------------------------------
- * Tcl_CmdtraceCmd --
+/*-----------------------------------------------------------------------------
+ * Tcl_CmdtraceObjCmd --
  *
  * Implements the TCL trace command:
  *     cmdtrace level|on ?noeval? ?notruncate? ?procs? ?fileid? ?command cmd?
@@ -391,26 +400,28 @@ CmdTraceRoutine (clientData, interp, level, command, cmdProc, cmdClientData,
  *-----------------------------------------------------------------------------
  */
 static int
-Tcl_CmdtraceCmd (clientData, interp, argc, argv)
-    ClientData    clientData;
-    Tcl_Interp   *interp;
-    int           argc;
-    char        **argv;
+TclX_CmdtraceObjCmd (clientData, interp, objc, objv)
+    ClientData  clientData;
+    Tcl_Interp *interp;
+    int         objc;
+    Tcl_Obj    *CONST objv[];
 {
     traceInfo_pt  infoPtr = (traceInfo_pt) clientData;
-    int           idx;
-    char         *fileHandle, *callback;
+    int idx;
+    char *argStr, *callback;
+    Tcl_Obj *channelId;
 
-    if (argc < 2)
+    if (objc < 2)
         goto argumentError;
+    argStr = Tcl_GetStringFromObj (objv [1], NULL);
 
     /*
      * Handle `depth' sub-command.
      */
-    if (STREQU (argv[1], "depth")) {
-        if (argc != 2)
+    if (STREQU (argStr, "depth")) {
+        if (objc != 2)
             goto argumentError;
-        sprintf(interp->result, "%d", infoPtr->depth);
+        Tcl_SetIntObj (Tcl_GetObjResult (interp),  infoPtr->depth);
         return TCL_OK;
     }
 
@@ -422,8 +433,8 @@ Tcl_CmdtraceCmd (clientData, interp, argc, argv)
     /*
      * Handle off sub-command.
      */
-    if (STREQU (argv[1], "off")) {
-        if (argc != 2)
+    if (STREQU (argStr, "off")) {
+        if (objc != 2)
             goto argumentError;
         return TCL_OK;
     }
@@ -432,99 +443,109 @@ Tcl_CmdtraceCmd (clientData, interp, argc, argv)
     infoPtr->noTruncate = FALSE;
     infoPtr->procCalls  = FALSE;
     infoPtr->channel    = NULL;
-    fileHandle          = NULL;
+    channelId           = NULL;
     callback            = NULL;
 
-    for (idx = 2; idx < argc; idx++) {
-        if (STREQU (argv[idx], "notruncate")) {
+    if (STREQU (argStr, "on")) {
+        infoPtr->depth = MAXINT;
+    } else {
+        if (Tcl_GetIntFromObj (interp, objv [1], &(infoPtr->depth)) != TCL_OK)
+            return TCL_ERROR;
+    }
+
+    for (idx = 2; idx < objc; idx++) {
+        argStr = Tcl_GetStringFromObj (objv [idx], NULL);
+        if (STREQU (argStr, "notruncate")) {
             if (infoPtr->noTruncate)
                 goto argumentError;
             infoPtr->noTruncate = TRUE;
             continue;
         }
-        if (STREQU (argv[idx], "noeval")) {
+        if (STREQU (argStr, "noeval")) {
             if (infoPtr->noEval)
                 goto argumentError;
             infoPtr->noEval = TRUE;
             continue;
         }
-        if (STREQU (argv[idx], "procs")) {
+        if (STREQU (argStr, "procs")) {
             if (infoPtr->procCalls)
                 goto argumentError;
             infoPtr->procCalls = TRUE;
             continue;
         }
-        if (STRNEQU (argv [idx], "std", 3) || 
-                STRNEQU (argv [idx], "file", 4)) {
-            if (fileHandle != NULL)
+        if (STRNEQU (argStr, "std", 3) || 
+                STRNEQU (argStr, "file", 4)) {
+            if (channelId != NULL)
                 goto argumentError;
             if (callback != NULL)
                 goto mixCommandAndFile;
-            fileHandle = argv [idx];
+            channelId = objv [idx];
             continue;
         }
-        if (STREQU (argv [idx], "command")) {
+        if (STREQU (argStr, "command")) {
             if (callback != NULL)
                 goto argumentError;
-            if (fileHandle != NULL)
+            if (channelId != NULL)
                 goto mixCommandAndFile;
-            if (idx == argc - 1)
+            if (idx == objc - 1)
                 goto missingCommand;
-            callback = argv [++idx];
+            callback = Tcl_GetStringFromObj (objv [++idx], NULL);
             continue;
         }
         goto invalidOption;
     }
 
-    if (STREQU (argv[1], "on")) {
-        infoPtr->depth = MAXINT;
-    } else {
-        if (Tcl_GetInt (interp, argv[1], &(infoPtr->depth)) != TCL_OK)
-            return TCL_ERROR;
-    }
-
     if (callback != NULL) {
         infoPtr->callback = ckstrdup (callback);
+        infoPtr->errorAsyncHandler =
+            Tcl_AsyncCreate (TraceCallbackErrorHandler, 
+                             (ClientData) infoPtr);
+
     } else {
-        if (fileHandle == NULL)
-            fileHandle = "stdout";
-        infoPtr->channel = TclX_GetOpenChannel (interp,
-                                                fileHandle,
-                                                TCL_WRITABLE);
+        if (channelId == NULL) {
+            infoPtr->channel = TclX_GetOpenChannel (interp,
+                                                    "stdout",
+                                                    TCL_WRITABLE);
+        } else {
+            infoPtr->channel = TclX_GetOpenChannelObj (interp,
+                                                       channelId,
+                                                       TCL_WRITABLE);
+        }
         if (infoPtr->channel == NULL)
             return TCL_ERROR;
     }
-    infoPtr->traceHolder = Tcl_CreateTrace (interp,
-                                            infoPtr->depth,
-                                            CmdTraceRoutine,
-                                            (ClientData) infoPtr);
+    infoPtr->traceId =
+        Tcl_CreateTrace (interp,
+                         infoPtr->depth,
+                         (Tcl_CmdTraceProc*) CmdTraceRoutine,
+                         (ClientData) infoPtr);
     return TCL_OK;
 
   argumentError:
-    Tcl_AppendResult (interp, tclXWrongArgs, argv [0], 
-                      " level | on ?noeval? ?notruncate? ?procs?",
-                      "?fileid? ?command cmd? | off | depth", (char *) NULL);
+    TclX_AppendObjResult (interp, tclXWrongArgs, objv [0], 
+                          " level | on ?noeval? ?notruncate? ?procs?",
+                          "?fileid? ?command cmd? | off | depth",
+                          (char *) NULL);
     return TCL_ERROR;
 
   missingCommand:
-    Tcl_AppendResult (interp, "command option requires an argument",
-                      (char *) NULL);
+    TclX_AppendObjResult (interp, "command option requires an argument",
+                          (char *) NULL);
     return TCL_ERROR;
 
   mixCommandAndFile:
-    Tcl_AppendResult (interp, "can not specify both the command option and ",
-                      "a file handle", (char *) NULL);
+    TclX_AppendObjResult (interp, "can not specify both the command option ",
+                          "and a file handle", (char *) NULL);
     return TCL_ERROR;
 
   invalidOption:
-    Tcl_AppendResult (interp, "invalid option: expected ",
-                      "one of \"noeval\", \"notruncate\", \"procs\", ",
-                      "\"command\", or a file id", (char *) NULL);
+    TclX_AppendObjResult (interp, "invalid option: expected ",
+                          "one of \"noeval\", \"notruncate\", \"procs\", ",
+                          "\"command\", or a file id", (char *) NULL);
     return TCL_ERROR;
 }
 
-/*
- *-----------------------------------------------------------------------------
+/*-----------------------------------------------------------------------------
  * DebugCleanUp --
  *
  *  Release the debug data area when the interpreter is deleted.
@@ -541,35 +562,40 @@ DebugCleanUp (clientData, interp)
     ckfree ((char *) infoPtr);
 }
 
-/*
- *-----------------------------------------------------------------------------
- * Tcl_InitDebug --
+/*-----------------------------------------------------------------------------
+ * TclX_DebugInit --
  *
  *  Initialize the TCL debugging commands.
  *-----------------------------------------------------------------------------
  */
 void
-Tcl_InitDebug (interp)
+TclX_DebugInit (interp)
     Tcl_Interp *interp;
 {
     traceInfo_pt infoPtr;
 
     infoPtr = (traceInfo_pt) ckalloc (sizeof (traceInfo_t));
 
-    infoPtr->interp      = interp;
-    infoPtr->traceHolder = NULL;
-    infoPtr->inTrace     = FALSE;
-    infoPtr->noEval      = FALSE;
-    infoPtr->noTruncate  = FALSE;
-    infoPtr->procCalls   = FALSE;
-    infoPtr->depth       = 0;
-    infoPtr->callback    = NULL;
-    infoPtr->channel     = NULL;
+    infoPtr->interp = interp;
+    infoPtr->traceId = NULL;
+    infoPtr->inTrace = FALSE;
+    infoPtr->noEval = FALSE;
+    infoPtr->noTruncate = FALSE;
+    infoPtr->procCalls = FALSE;
+    infoPtr->depth = 0;
+    infoPtr->callback = NULL;
+    infoPtr->errorStatePtr = NULL;
+    infoPtr->errorAsyncHandler = NULL;
+    infoPtr->channel = NULL;
 
     Tcl_CallWhenDeleted (interp, DebugCleanUp, (ClientData) infoPtr);
 
-    Tcl_CreateCommand (interp, "cmdtrace", Tcl_CmdtraceCmd, 
-                       (ClientData) infoPtr, (void (*)()) NULL);
+    Tcl_CreateObjCommand (interp, "cmdtrace",
+                          TclX_CmdtraceObjCmd, 
+                          (ClientData) infoPtr,
+                          (Tcl_CmdDeleteProc*) NULL);
 }
 
 
+
+
